# -*- coding:utf-8 -*-
import time
import zlib
import ctypes
import inspect
import baostock.data.resultset as rs
import baostock.common.contants as cons
import baostock.util.stringutil as strUtil
import baostock.common.context as conx
import baostock.util.socketutil as sock
import baostock.data.messageheader as msgheader


def subscribe_by_code(code_list, subscribe_type=0, fncallback=None, options="", user_params=None):
    """订阅实时行情
    @param code_list: 每只证券代码之间用“英文逗号分隔符”，结尾不存在“英文逗号分隔符”；证券代码格式：sh.600000，前两位为证券市场：“sz”深圳、“sh”上海
    @param subscribe_type: 订阅方式 0：按证券代码订阅， 1：按行情数据类型订阅。0.7.5版本只支持按证券代码订阅；
    @param fncallback: 自定义回调方法
    @param options: 预留参数
    @param user_params: 用户参数，回调时原样返回
    """
    data = rs.SubscibeData()
    data.method = "subscribe_by_code"
    data.options = options
    data.user_params = user_params

    if code_list is None or code_list == "":
        data.error_code = cons.BSERR_CODE_INVALIED
        data.error_msg = "证券代码不正确"
        return data

    # 判断code_list，如果大于限制，则进行截取
    if len(code_list.split(cons.ATTRIBUTE_SPLIT)) > cons.BAOSTOCK_REALTIME_LIMIT_COUNT:
        code_list = cons.ATTRIBUTE_SPLIT.join(code_list.split(cons.ATTRIBUTE_SPLIT)[0:cons.BAOSTOCK_REALTIME_LIMIT_COUNT])

    data.code_list = code_list.lower()

    # 设置订阅方式 默认为0
    # if subscribe_type == 0 or subscribe_type == 1:
    if subscribe_type == 0:
        pass
    else:
        subscribe_type = 0
    data.subscribe_type = subscribe_type

    if not callable(fncallback):
        fncallback = DemoCallback
    data.fncallback = fncallback
    user_id = ""
    try:
        user_id = getattr(conx, "user_id")
    except Exception:
        print("you don't login.")
        data.error_code = cons.BSERR_NO_LOGIN
        data.error_msg = "you don't login."
        return data

    data.user_id = user_id
    param = "%s\1%s\1%s\1%s\1%s\1%s\1%s" % (
        "subscribe_by_code", user_id, subscribe_type,
        code_list, fncallback, options, user_params)
    msg_body = strUtil.organize_realtime_msg_body(param)
    msg_header = msgheader.to_message_header(
        cons.MESSAGE_TYPE_SUBSCRIPTIONS_BY_SECURITYID_REQUEST, len(msg_body))
    data.msg_type = cons.MESSAGE_TYPE_SUBSCRIPTIONS_BY_SECURITYID_REQUEST
    data.msg_body = msg_body

    head_body = msg_header + msg_body
    crc32str = zlib.crc32(bytes(head_body, encoding='utf-8'))

    sock.send_real_time_subscibe(head_body + cons.MESSAGE_SPLIT + str(crc32str), data)

    return data


def cancel_subscribe(ident):
    """取消实时行情
    @param thread: 实时行情的线程；
    """
    data = rs.SubscibeData()

    # 取订阅的子循环
    try:
        _async_raise(ident, SystemExit)
    except (ValueError, TypeError):
        data.error_code = cons.BSERR_PARAM_ERR
        data.error_msg = "取消订阅失败"
        return data

    # 封装待发出的消息
    user_id = ""
    try:
        user_id = getattr(conx, "user_id")
    except Exception:
        print("you don't login.")
        data.error_code = cons.BSERR_NO_LOGIN
        data.error_msg = "you don't login."
        return data

    data.user_id = user_id

    # 组织体信息
    msg_body = "cancel_subscribe" + cons.MESSAGE_SPLIT + user_id

    # 组织头信息
    msg_header = msgheader.to_message_header(
        cons.MESSAGE_TYPE_CANCEL_SUBSCRIBE_REQUEST, len(msg_body))
    head_body = msg_header + msg_body

    crc32str = zlib.crc32(bytes(head_body, encoding='utf-8'))

    # 向服务器发送取消订阅
    sock.send_cancel_real_time_msg(head_body + cons.MESSAGE_SPLIT + str(crc32str))

    data.error_code = cons.BSERR_SUCCESS
    data.error_msg = "success"

#     if receive_data is None or receive_data.strip() == "":
#         data.error_code = cons.BSERR_RECVSOCK_FAIL
#         data.error_msg = "网络接收错误。"
#         return data
# 
#     msg_header = receive_data[0:cons.MESSAGE_HEADER_LENGTH]
#     msg_body = receive_data[cons.MESSAGE_HEADER_LENGTH:-1]
# 
#     header_arr = msg_header.split(cons.MESSAGE_SPLIT)
#     body_arr = msg_body.split(cons.MESSAGE_SPLIT)
# 
#     data.msg_type = header_arr[1]
#     data.msg_body_length = header_arr[2]
# 
#     data.error_code = body_arr[0]
#     data.error_msg = body_arr[1]

    return data


def _async_raise(tid, exctype):
    """raises the exception, performs cleanup if needed"""
    tid = ctypes.c_long(tid)
    if not inspect.isclass(exctype):
        exctype = type(exctype)
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, ctypes.py_object(exctype))
    if res == 0:
        raise ValueError("invalid thread id")
    elif res != 1:
        ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, None)
        raise SystemError("PyThreadState_SetAsyncExc failed")


def DemoCallback(quantdata):
    """
    DemoCallback 是订阅时提供的回调函数模板。该函数只有一个为ResultData类型的参数quantdata
    :param quantdata:ResultData
    :return:
    """
    print(quantdata.data)



import baostock as bs
import pandas as pd

#### 登陆系统 ####
lg = bs.login()
# 显示登陆返回信息
print('login respond error_code:'+lg.error_code)
print('login respond  error_msg:'+lg.error_msg)

#### 获取公司业绩预告 ####
rs_forecast = bs.query_forecast_report("sh.600000", start_date="2010-01-01", end_date="2017-12-31")
print('query_forecast_reprot respond error_code:'+rs_forecast.error_code)
print('query_forecast_reprot respond  error_msg:'+rs_forecast.error_msg)
rs_forecast_list = []
while (rs_forecast.error_code == '0') & rs_forecast.next():
    # 分页查询，将每页信息合并在一起
    rs_forecast_list.append(rs_forecast.get_row_data())
result_forecast = pd.DataFrame(rs_forecast_list, columns=rs_forecast.fields)
#### 结果集输出到csv文件 ####
result_forecast.to_csv("D:\\forecast_report.csv", encoding="gbk", index=False)
print(result_forecast)

#### 登出系统 ####
bs.logout()



import baostock as bs
import pandas as pd

# 登陆系统
lg = bs.login()
# 显示登陆返回信息
print('login respond error_code:'+lg.error_code)
print('login respond  error_msg:'+lg.error_msg)

# 查询2015至2017年复权因子
rs_list = []
rs_factor = bs.query_adjust_factor(code="sh.600000", start_date="2015-01-01", end_date="2017-12-31")
while (rs_factor.error_code == '0') & rs_factor.next():
    rs_list.append(rs_factor.get_row_data())
result_factor = pd.DataFrame(rs_list, columns=rs_factor.fields)
# 打印输出
print(result_factor)

# 结果集输出到csv文件
result_factor.to_csv("D:\\adjust_factor_data.csv", encoding="gbk", index=False)

# 登出系统
bs.logout()



# -*- coding:utf-8 -*-
"""
获取默认socket
@author: baostock.com
@group : baostock.com
@contact: baostock@163.com
"""
import time
import socket
import threading
import zlib
import baostock.common.contants as cons
import baostock.common.context as context


class SocketUtil(object):
    """Socket工具类"""
    # 记录第一个被创建对象的引用
    instance = None
    # 记录是否执行过初始化动作
    init_flag = False

    def __new__(cls, *args, **kwargs):
        # 1. 判断类属性是否是空对象
        if cls.instance is None:
            # 2. 调用父类的方法，为第一个对象分配空间
            cls.instance = super().__new__(cls)
        # 3. 返回类属性保存的对象引用
        return cls.instance

    def __init__(self):
        SocketUtil.init_flag = True

    def connect(self):
        """创建连接"""
        try:
            mySockect = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            mySockect.connect((cons.BAOSTOCK_SERVER_IP, cons.BAOSTOCK_SERVER_PORT))
        except Exception:
            print("服务器连接失败，请稍后再试。")
        setattr(context, "default_socket", mySockect)


def get_default_socket():
    """获取默认连接"""
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((cons.BAOSTOCK_SERVER_IP, cons.BAOSTOCK_SERVER_PORT))
    except Exception:
        print("服务器连接失败，请稍后再试。")
        return None
    return sock


def send_msg(msg):
    """发送消息，并接受消息 """
    try:
        # default_socket = get_default_socket()
        if hasattr(context, "default_socket"):
            default_socket = getattr(context, "default_socket")
            if default_socket is not None:
                # str 类型 -> bytes 类型
                # msg = msg + "<![CDATA[]]>"  # 在消息结尾追加“消息之间的分隔符”，压缩时的分隔符
                msg = msg + "\n"  # 在消息结尾追加“消息之间的分隔符”，不压缩时的分隔符
                default_socket.send(bytes(msg, encoding='utf-8'))
                receive = b""
                while True:
                    recv = default_socket.recv(8192)
                    receive += recv
                    # 判断是否读取完
                    if receive[-13:] == b"<![CDATA[]]>\n":  # 压缩时的结尾分隔符长度
                    # if receive[-1:] == b"\n":  # 不压缩时的结尾分隔符长度
                        break
                # return bytes.decode(zlib.decompress(receive))  # 进行解压
                head_bytes = receive[0:cons.MESSAGE_HEADER_LENGTH]
                head_str = bytes.decode(head_bytes)
                head_arr = head_str.split(cons.MESSAGE_SPLIT)
                if head_arr[1] in cons.COMPRESSED_MESSAGE_TYPE_TUPLE:
                    # 消息体需要解压
                    head_inner_length = int(head_arr[2])
                    body_str = bytes.decode(zlib.decompress(receive[cons.MESSAGE_HEADER_LENGTH:cons.MESSAGE_HEADER_LENGTH + head_inner_length]))
                    return head_str + body_str
                else:
                    return bytes.decode(receive)  # 不进行解压
            else:
                return None
        else:
            print("you don't login.")
    except Exception as ex:
        print(ex)
        print("接收数据异常，请稍后再试。")


class SocketRealTimeUtil(object):
    """Socket实时工具类"""
    # 记录第一个被创建对象的引用
    instance = None
    # 记录是否执行过初始化动作
    init_flag = False

    def __new__(cls, *args, **kwargs):
        # 1. 判断类属性是否是空对象
        if cls.instance is None:
            # 2. 调用父类的方法，为第一个对象分配空间
            cls.instance = super().__new__(cls)
        # 3. 返回类属性保存的对象引用
        return cls.instance

    def __init__(self):
        SocketRealTimeUtil.init_flag = True

    def connect(self):
        """创建实时行情连接"""
        try:
            mySockect = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            mySockect.connect((cons.BAOSTOCK_SERVER_REAL_TIME_IP, cons.BAOSTOCK_SERVER_REAL_TIME_PORT))
        except Exception:
            print("服务器连接失败，请稍后再试。")
        setattr(context, "socket_real_time", mySockect)


def get_real_time_socket():
    """获取实时行情连接"""
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((cons.BAOSTOCK_SERVER_REAL_TIME_IP, cons.BAOSTOCK_SERVER_REAL_TIME_PORT))
    except Exception:
        print("服务器连接失败，请稍后再试。")
        return None
    return sock


def send_real_time_msg(msg):
    """发送实时行情消息，并接受消息 """
    try:
        # default_socket = get_default_socket()
        if hasattr(context, "socket_real_time"):
            default_socket = getattr(context, "socket_real_time")
            if default_socket is not None:
                # str 类型 -> bytes 类型
                msg = msg + "<![CDATA[]]>"  # 在消息结尾追加“消息之间的分隔符”
                default_socket.send(bytes(msg, encoding='utf-8'))
                if msg.split(cons.MESSAGE_SPLIT)[1] == cons.MESSAGE_TYPE_LOGOUT_REAL_TIME_REQUEST:
                    default_socket.close()
                    setattr(context, "default_socket", None)
                else:
                    # default_socket.close()
                    receive = b""
                    while True:
                        recv = default_socket.recv(8192)
                        receive += recv
                        # 判断是否读取完
                        if receive[-12:] == b"<![CDATA[]]>":
                            break
                    # return bytes.decode(receive)
                    return bytes.decode(zlib.decompress(receive))
            else:
                return None
        else:
            print("you don't login.")
    except Exception as ex:
        print(ex)
        print("接收数据异常，请稍后再试。")


def send_cancel_real_time_msg(msg):
    """发送取消订阅实时行情消息 """
    try:
        # default_socket = get_default_socket()
        if hasattr(context, "socket_real_time"):
            default_socket = getattr(context, "socket_real_time")

            if default_socket is not None:
                # str 类型 -> bytes 类型
                msg = msg + "<![CDATA[]]>"  # 在消息结尾追加“消息之间的分隔符”
                default_socket.send(bytes(msg, encoding='utf-8'))
            else:
                return None
        else:
            print("you don't login.")
    except Exception as ex:
        print(ex)
        print("接收数据异常，请稍后再试。")


def real_time_subscibe_thread(receive, socket_real_time, data):
    # time.sleep(2)
    # receive = b""
    # data.fncallback(data)
    data.data = {}
    while True:
        try:
            recv = socket_real_time.recv(8192)
        except ConnectionAbortedError:
            # 如果连接被断开，跳出循环，不再接收数据
            break
        receive += recv
        # print(receive)
        # 判断是否读取完b"<![CDATA[]]>"
        if receive.find(b"<![CDATA[]]>"):

            if receive[-12:] == b"<![CDATA[]]>":
                # 如果接收的数据以'<![CDATA[]]>'结尾，则是一个完整的记录，
                recv_arr = receive.split(b"<![CDATA[]]>")
                receive = b""
            else:
                recv_arr = receive.split(b"<![CDATA[]]>")
                recv_arr = receive.split(b"<![CDATA[]]>")[0:(len(receive.split(b"<![CDATA[]]>")) - 1)]
                receive = receive.split(b"<![CDATA[]]>")[-1:][0]

            for real_time_byte in recv_arr:
                # 由于real_time_byte可能为b''，即为空，需要判断
                if len(real_time_byte) > 0:
                    try:
                        real_time = bytes.decode(zlib.decompress(real_time_byte))
                    except Exception as ex:
                        print(ex)
                        print("数据解压缩异常，请稍后再试或联系管理员。")
                    if real_time is None or real_time.strip() == "":
                        continue

                    # print(real_time)

                    if real_time.find("<![CDATA[]]>"):
                        real_time_arr = real_time.split("<![CDATA[]]>")
                        for realtime_arr_item in real_time_arr:
                            msg_header = realtime_arr_item[0:cons.MESSAGE_HEADER_LENGTH]
                            msg_header_arr = msg_header.split(cons.MESSAGE_SPLIT)
                            msg_body_length = int(msg_header_arr[2])
                            msg_body = realtime_arr_item[cons.MESSAGE_HEADER_LENGTH:cons.MESSAGE_HEADER_LENGTH + msg_body_length]

                            receive_crc = realtime_arr_item[cons.MESSAGE_HEADER_LENGTH + msg_body_length:]  # 传递进来的CRC校验码
                            crc32str = zlib.crc32(bytes(msg_header+msg_body, "utf8"))  # 客户端计算的CRC校验码
                            if receive_crc == str(crc32str):

                                body_arr = msg_body.split(cons.MESSAGE_SPLIT)
                                try:
                                    data.error_code = body_arr[0]
                                    data.error_msg = body_arr[1]

                                    if cons.BSERR_SUCCESS == data.error_code:
                                        realtime_arr_item = body_arr[3]
                                        data.data[realtime_arr_item.split(",")[2]] = realtime_arr_item.split(',')
                                        data.error_code = data.error_code  # 错误代码
                                        data.error_msg = data.error_msg
                                        data.fncallback(data)
                                    # data.error_code = cons.BSERR_NO_LOGIN  # 错误代码
                                    # data.error_msg = ""
                                    data.data = {}
                                except IndexError:
                                    pass

                    else:
                        msg_header = real_time[0:cons.MESSAGE_HEADER_LENGTH]
                        msg_header_arr = msg_header.split(cons.MESSAGE_SPLIT)
                        msg_body_length = int(msg_header_arr[2])
                        msg_body = real_time[cons.MESSAGE_HEADER_LENGTH:cons.MESSAGE_HEADER_LENGTH + msg_body_length]

                        receive_crc = real_time[cons.MESSAGE_HEADER_LENGTH + msg_body_length:]  # 传递进来的CRC校验码
                        crc32str = zlib.crc32(bytes(msg_header+msg_body, "utf8"))  # 客户端计算的CRC校验码
                        if receive_crc == str(crc32str):

                            body_arr = msg_body.split(cons.MESSAGE_SPLIT)
                            try:
                                data.error_code = body_arr[0]
                                data.error_msg = body_arr[1]

                                if cons.BSERR_SUCCESS == data.error_code:
                                    real_time = body_arr[3]
                                    data.data[real_time.split(",")[2]] = real_time.split(',')
                                    data.error_code = data.error_code  # 错误代码
                                    data.error_msg = data.error_msg
                                    data.fncallback(data)
                                # data.error_code = cons.BSERR_NO_LOGIN  # 错误代码
                                # data.error_msg = ""
                                data.data = {}
                            except IndexError:
                                pass
                else:
                    pass
#                     print(real_time_byte)
#                     print(bytes.decode(real_time_byte))


def send_real_time_subscibe(msg, data):
    """发送实时行情消息，并接受消息 """
    try:
        # default_socket = get_default_socket()
        if hasattr(context, "socket_real_time"):
            socket_real_time = getattr(context, "socket_real_time")
            if socket_real_time is not None:
                # str 类型 -> bytes 类型
                msg = msg + "<![CDATA[]]>"  # 在消息结尾追加“消息之间的分隔符”
                socket_real_time.send(bytes(msg, encoding='utf-8'))
                receive = b""
                while True:
                    recv = socket_real_time.recv(8192)
                    receive += recv
                    # 判断是否读取完
                    if receive[-12:] == b"<![CDATA[]]>":
                        # receive_data = bytes.decode(receive)
                        """ 解压  """
                        try:
                            receive_data = bytes.decode(zlib.decompress(receive[:-12]))
                        except Exception as ex:
                            receive = b""
                            print(ex)
                            print("数据解压缩异常，请稍后再试或联系管理员。")
                            continue

                        if receive_data is None or receive_data.strip() == "":
                            data.error_code = cons.BSERR_RECVSOCK_FAIL
                            data.error_msg = "网络接收错误。"
                            return data

                        msg_header = receive_data[0:cons.MESSAGE_HEADER_LENGTH]
                        msg_header_arr = msg_header.split(cons.MESSAGE_SPLIT)
                        msg_body_length = msg_header_arr[2]
                        msg_body = receive_data[cons.MESSAGE_HEADER_LENGTH:cons.MESSAGE_HEADER_LENGTH + int(msg_body_length)]

                        # header_arr = msg_header.split(cons.MESSAGE_SPLIT)
                        body_arr = msg_body.split(cons.MESSAGE_SPLIT)

                        data.error_code = body_arr[0]
                        data.error_msg = body_arr[1]

                        if cons.BSERR_SUCCESS == data.error_code:
                            # real_time = body_arr[3]
                            # data.data[real_time.split(",")[0]] = real_time.strip(',').split(',')
                            t = threading.Thread(target=real_time_subscibe_thread, args=(receive, socket_real_time, data))
                            t.start()
                            data.serial_id = t.ident
                        data.error_code = data.error_code  # 错误代码
                        data.error_msg = data.error_msg
                        break
                return data
            else:
                return None
        else:
            print("you don't login.")
    except Exception as ex:
        print(ex)
        print("接收数据异常，请稍后再试。")



import baostock as bs
import pandas as pd

# 登陆系统
lg = bs.login()
# 显示登陆返回信息
print('login respond error_code:'+lg.error_code)
print('login respond  error_msg:'+lg.error_msg)

# 季频现金流量
cash_flow_list = []
rs_cash_flow = bs.query_cash_flow_data(code="sh.600000", year=2017, quarter=2)
while (rs_cash_flow.error_code == '0') & rs_cash_flow.next():
    cash_flow_list.append(rs_cash_flow.get_row_data())
result_cash_flow = pd.DataFrame(cash_flow_list, columns=rs_cash_flow.fields)
# 打印输出
print(result_cash_flow)
# 结果集输出到csv文件
result_cash_flow.to_csv("D:\\cash_flow_data.csv", encoding="gbk", index=False)

# 登出系统
bs.logout()



import baostock as bs
import pandas as pd

#### 登陆系统 ####
lg = bs.login()
# 显示登陆返回信息
print('login respond error_code:'+lg.error_code)
print('login respond  error_msg:'+lg.error_msg)

#### 查询除权除息信息####
# 查询2015年除权除息信息
rs_list = []
rs_dividend_2015 = bs.query_dividend_data(code="sh.600000", year="2015", yearType="report")
while (rs_dividend_2015.error_code == '0') & rs_dividend_2015.next():
    rs_list.append(rs_dividend_2015.get_row_data())

# 查询2016年除权除息信息
rs_dividend_2016 = bs.query_dividend_data(code="sh.600000", year="2016", yearType="report")
while (rs_dividend_2016.error_code == '0') & rs_dividend_2016.next():
    rs_list.append(rs_dividend_2016.get_row_data())

# 查询2017年除权除息信息
rs_dividend_2017 = bs.query_dividend_data(code="sh.600000", year="2017", yearType="report")
while (rs_dividend_2017.error_code == '0') & rs_dividend_2017.next():
    rs_list.append(rs_dividend_2017.get_row_data())

result_dividend = pd.DataFrame(rs_list, columns=rs_dividend_2017.fields)
# 打印输出
print(result_dividend)

#### 结果集输出到csv文件 ####   
result_dividend.to_csv("D:\\history_Dividend_data.csv", encoding="gbk",index=False)

#### 登出系统 ####
bs.logout()



import baostock as bs
import pandas as pd

#### 登陆系统 ####
lg = bs.login()
# 显示登陆返回信息
print('login respond error_code:'+lg.error_code)
print('login respond  error_msg:'+lg.error_msg)

#### 获取证券信息 ####
rs = bs.query_all_stock(day="2017-06-30")
print('query_all_stock respond error_code:'+rs.error_code)
print('query_all_stock respond  error_msg:'+rs.error_msg)

#### 打印结果集 ####
data_list = []
while (rs.error_code == '0') & rs.next():
    # 获取一条记录，将记录合并在一起
    data_list.append(rs.get_row_data())
result = pd.DataFrame(data_list, columns=rs.fields)

#### 结果集输出到csv文件 ####   
result.to_csv("D:\\all_stock.csv", encoding="gbk", index=False)
print(result)

#### 登出系统 ####
bs.logout()



import baostock as bs
import pandas as pd

# 登陆系统
lg = bs.login()
# 显示登陆返回信息
print('login respond error_code:'+lg.error_code)
print('login respond  error_msg:'+lg.error_msg)

# 获取终止上市股票列表
rs = bs.query_terminated_stocks('2018-09-12')
print('query_terminated error_code:'+rs.error_code)
print('query_terminated  error_msg:'+rs.error_msg)

# 打印结果集
terminated_stocks = []
while (rs.error_code == '0') & rs.next():
    # 获取一条记录，将记录合并在一起
    terminated_stocks.append(rs.get_row_data())
result = pd.DataFrame(terminated_stocks, columns=rs.fields)
# 结果集输出到csv文件
result.to_csv("D:/terminated_stocks.csv", encoding="gbk", index=False)
print(result)

# 登出系统
bs.logout()
